<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace TamerLib\Classes;

    use InvalidArgumentException;
    use LogLib\Log;
    use RuntimeException;
    use TamerLib\Exceptions\WorkerFailedException;
    use TamerLib\Objects\ServerConfiguration;
    use TamerLib\Objects\WorkerConfiguration;
    use TamerLib\Objects\WorkerInstance;

    class WorkerSupervisor
    {
        /**
         * @var WorkerInstance[]
         */
        private $workers;

        /**
         * @var ServerConfiguration
         */
        private $configuration;

        /**
         * WorkerSupervisor constructor.
         */
        public function __construct(ServerConfiguration $configuration)
        {
            $this->workers = [];
            $this->configuration = $configuration;
        }

        /**
         * Generates a worker configuration object.
         *
         * @param int $channel The channel to use for the worker.
         * @return WorkerConfiguration The generated worker configuration.
         */
        private function generateWorkerConfiguration(int $channel=0): WorkerConfiguration
        {
            $configuration = new WorkerConfiguration();
            $configuration->setHost($this->configuration->getHost());
            $configuration->setPort($this->configuration->getPort());
            $configuration->setPassword($this->configuration->getPassword());
            $configuration->setDatabase($this->configuration->getDatabase());
            $configuration->setChannel($channel);

            return $configuration;
        }

        /**
         * Spawns a specified number of workers for a given path.
         *
         * @param string $path Optional. The path to the worker file, defaults to subproc.
         * @param int $count Optional. The number of workers to spawn, defaults to 8.
         * @param int $channel Optional. The channel to use for the workers, defaults to 0.
         * @param bool $check Optional. Whether to check if the workers have started, defaults to true.
         * @throws WorkerFailedException Thrown if the worker fails to start.
         * @return void
         */
        public function spawnWorker(string $path, int $count=8, int $channel=0, bool $check=true): void
        {
            if(!file_exists($path) || !is_file($path))
            {
                throw new InvalidArgumentException(sprintf('Path %s does not exist', $path));
            }

            Log::verbose(Utilities::getName(), sprintf('Spawning %s workers for %s', $count, $path));
            $spawned_workers = [];

            for($i = 0; $i < $count; $i++)
            {
                $worker_config = $this->generateWorkerConfiguration($channel);
                $worker = new WorkerInstance($worker_config, $path);
                $this->workers[$worker_config->getWorkerId()] = $worker;
                $spawned_workers[$worker_config->getWorkerId()] = time();
                $worker->start();
            }

            if($check)
            {
                $this->checkWorkers($spawned_workers);
            }
        }

        /**
         * Checks if the workers have started.
         *
         * @param array $workers The workers to check.
         * @throws WorkerFailedException Thrown if the worker fails to start.
         * @return void
         */
        private function checkWorkers(array $workers): void
        {
            while(true)
            {
                if(count($workers) === 0)
                {
                    return;
                }

                foreach($workers as $worker_id => $time)
                {
                    if(time() - $time > 3)
                    {
                        if($this->workers[$worker_id]->getProcess()?->isRunning() === false)
                        {
                            throw new WorkerFailedException(sprintf('Worker %s failed, has not started in %s seconds', $worker_id, 1));
                        }

                        Log::debug(Utilities::getName(), sprintf('Worker %s has started in %s seconds', $worker_id, 1));
                        unset($workers[$worker_id]);
                    }
                }

                $this->printUpdates();
            }
        }

        /**
         * Spawns a closure as a worker process. (built-in worker)
         *
         * @param int $count The number of workers to spawn.
         * @param int $channel The channel to use for the workers.
         * @throws WorkerFailedException Thrown if the worker fails to start.
         * @return void
         */
        public function spawnClosure(int $count=8, int $channel=0): void
        {
            if(!file_exists(__DIR__ . DIRECTORY_SEPARATOR . 'subproc'))
            {
                throw new RuntimeException(sprintf('subproc file does not exist, checked %s', __DIR__));
            }

            $this->spawnWorker(__DIR__ . DIRECTORY_SEPARATOR . 'subproc', $count, $channel);
        }

        /**
         * @return void
         */
        public function printUpdates(): void
        {
            if(count($this->workers) === 0)
            {
                return;
            }

            /** @var WorkerInstance $worker */
            foreach($this->workers as $worker)
            {
                print($worker->getOutput());
            }
        }

        /**
         * Monitors the worker processes and restarts them if they are not running.
         *
         * @param int $timeout
         * @return void
         * @throws WorkerFailedException
         */
        public function monitor(int $timeout=0): void
        {
            if(count($this->workers) === 0)
            {
                return;
            }

            $start_time = time();

            while(true)
            {
                /** @var WorkerInstance $worker */
                foreach($this->workers as $worker)
                {
                    print($worker->getOutput());

                    if(!$worker->isRunning())
                    {
                        Log::warning(Utilities::getName(), sprintf('Worker %s is not running, killing', $worker->getConfiguration()->getWorkerId()));
                        $worker->restart();
                    }
                }

                if($timeout < 0)
                {
                    return;
                }

                if($timeout > 0 && time() - $start_time > $timeout)
                {
                    return;
                }
            }

        }

        /**
         * Stops all worker processes and removes them from the supervisor.
         *
         * @return void
         */
        public function stopAll(): void
        {
            if(count($this->workers) === 0)
            {
                return;
            }

            Log::verbose(Utilities::getName(), 'Stopping all workers');
            foreach($this->workers as $worker_id => $worker)
            {
                $worker->stop();
                unset($this->workers[$worker_id]);
            }
        }

        /**
         * Public Destructor
         */
        public function __destruct()
        {
            $this->stopAll();
        }
    }